/* based on: $XConsortium: xload.c,v 1.37 94/04/17 20:43:44 converse Exp $ */
/*

Copyright (c) 1989  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/

/*
 * xpppload - display ISDN load average in a window
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <linux/isdn.h>

#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/StripChart.h>
#include <X11/Xmu/SysUtil.h>

#ifndef STREAMS
#if defined(linux) && defined(__powerpc__) \
    && (__GLIBC__ == 2 && __GLIBC_MINOR__ == 0)
/* kludge alert! */
#undef __GLIBC__
#endif
#include <sys/socket.h>		/* *BSD, Linux, NeXT, Ultrix etc. */
#ifndef linux
#include <net/if.h>
#include <net/ppp_defs.h>
#include <net/if_ppp.h>
#else
/* Linux */
#if __GLIBC__ >= 2
#include <asm/types.h>		/* glibc 2 conflicts with linux/types.h */
#include <net/if.h>
#else
#include <linux/types.h>
#include <linux/if.h>
#endif
#include <linux/ppp_defs.h>
#include <linux/if_ppp.h>
#endif /* linux */

#else	/* STREAMS */
#include <sys/stropts.h>	/* SVR4, Solaris 2, SunOS 4, OSF/1, etc. */
#include <net/ppp_defs.h>
#include <net/pppio.h>

#endif	/* STREAMS */

#include "xpppload.bit"

char *ProgramName;

static void quit();

/*
 * Definition of the Application resources structure.
 */

typedef struct _XPPPLoadResources {
  Boolean show_label;
  char *online_color;
  char *active_color;
  char *trying_color;
  char *activate;
  char *deactivate;
} XPPPLoadResources;

/*
 * Command line options table.  Only resources are entered here...there is a
 * pass over the remaining options after XtParseCommand is let loose.
 */

static XrmOptionDescRec options[] = {
 {"-scale",		"*load.minScale",	XrmoptionSepArg,	NULL},
 {"-update",		"*load.update",		XrmoptionSepArg,	NULL},
 {"-hl",		"*load.highlight",	XrmoptionSepArg,	NULL},
 {"-highlight",		"*load.highlight",	XrmoptionSepArg,	NULL},
 {"-label",		"*label.label",		XrmoptionSepArg,	NULL},
 {"-nolabel",		"*showLabel",	        XrmoptionNoArg,       "False"},
 {"-jumpscroll",	"*load.jumpScroll",	XrmoptionSepArg,	NULL},
 {"-online",		"*onlineColor",         XrmoptionSepArg,	NULL},
 {"-trying",		"*tryingColor",         XrmoptionSepArg,	NULL},
 {"-active",		"*activeColor",         XrmoptionSepArg,	NULL},
 {"-activate",		"*activate",		XrmoptionSepArg,	NULL},
 {"-deactivate",	"*deactivate",		XrmoptionSepArg,	NULL},
};

/*
 * The structure containing the resource information for the
 * Xpppload application resources.
 */

#define Offset(field) (XtOffsetOf(XPPPLoadResources, field))

static XtResource my_resources[] = {
  {"showLabel", XtCBoolean, XtRBoolean, sizeof(Boolean),
     Offset(show_label), XtRImmediate, (XtPointer) TRUE},
  {"onlineColor", "OnlineColor", XtRString, sizeof(char *),
     Offset(online_color), XtRString, NULL},
  {"activeColor", "ActiveColor", XtRString, sizeof(char *),
     Offset(active_color), XtRString, NULL},
  {"tryingColor", "TryingColor", XtRString, sizeof(char *),
     Offset(trying_color), XtRString, NULL},
  {"activate", "Activate", XtRString, sizeof(char *),
     Offset(activate), XtRString, NULL},
  {"deactivate", "Deactivate", XtRString, sizeof(char *),
     Offset(deactivate), XtRString, NULL},
};

#undef Offset

static XPPPLoadResources resources;

static XtActionsRec xpppload_actions[] = {
    { "quit",	quit },
};
static Atom wm_delete_window;

static char *IfName=NULL;

#define STATE_OFFLINE	0
#define STATE_ONLINE	1
#define STATE_ACTIVE	2
#define STATE_TRYING	3	/* currently not supported :-( I habe
                                   no idea, how to figure it out. */


static int	State=STATE_OFFLINE;
static time_t	OnlineSince=0;
static time_t	OnlineTill=0;
static time_t	Now;
static int	ByteCount=0;

static Pixel onlinecolor, activecolor, tryingcolor, bgcolor;

static int interface_active=0;
static Widget label_wid;

static int fd_info;
static struct ppp_stats cur;

/*
 * Exit with message describing command line format.
 */

void usage()
{
    fprintf (stderr, "usage:  %s [-options ...] ifname\n\n", ProgramName);
    fprintf (stderr, "where options include:\n");
    fprintf (stderr,
      "    -display dpy            X server on which to display\n");
    fprintf (stderr,
      "    -geometry geom          size and location of window\n");
    fprintf (stderr,
      "    -fn font                font to use in label\n");
    fprintf (stderr,
      "    -scale number           minimum number of scale lines\n");
    fprintf (stderr,
      "    -update seconds         interval between updates\n");
    fprintf (stderr,
      "    -label string           annotation text\n");
    fprintf (stderr,
      "    -bg color               background color\n");
    fprintf (stderr,
      "    -fg color               graph color\n");
    fprintf (stderr,
      "    -hl color               scale and text color\n");
    fprintf (stderr,
      "    -online color           background color when online\n");
    fprintf (stderr,
      "    -active color           background color when active for demand dialing\n");
    fprintf (stderr,
      "    -nolabel                removes the label from above the chart.\n");
    fprintf (stderr,
      "    -jumpscroll value       number of pixels to scroll on overflow\n");
    fprintf (stderr,
      "    -activate string        exec this to activate demand dialing\n");
    fprintf (stderr,
      "    -deactivate string      exec this to deactivate demand dialing\n");
    fprintf (stderr, "\n");
    exit(1);
}



int get_active()
{
  struct ifreq ifr;
  int	mtu;

  strcpy(ifr.ifr_name, IfName);
  if (ioctl(fd_info, SIOCGIFMTU, &ifr) < 0)
    return STATE_OFFLINE;
  mtu=ifr.ifr_mtu;

  if (ioctl(fd_info, SIOCGIFFLAGS, &ifr) < 0)
    return STATE_OFFLINE;

  if (ifr.ifr_flags&IFF_UP)
    if (mtu <1500)		/* FIXME: only valid for PPPoE */
      return STATE_ONLINE;
    else
      return STATE_ACTIVE;

  return STATE_OFFLINE;
}

#ifndef STREAMS
static void
get_ppp_stats(curp)
     struct ppp_stats *curp;
{
  struct ifpppstatsreq req;

  memset (&req, 0, sizeof (req));

#ifdef linux
  req.stats_ptr = (caddr_t) &req.stats;
#undef ifr_name
#define ifr_name ifr__name
#endif

  strncpy(req.ifr_name, IfName, sizeof(req.ifr_name));
  if (ioctl(fd_info, SIOCGPPPSTATS, &req) < 0)
    {
      if (errno != ENODEV)
	{
	  fprintf(stderr, "%s: ", ProgramName);
	  if (errno == ENOTTY)
	    fprintf(stderr, "kernel support missing\n");
	  else
	    perror("couldn't get PPP statistics");
	  exit(1);
	}
    }
  else
    {
      interface_active=1;
      *curp = req.stats;
    }
}

#else	/* STREAMS */

int
strioctl(fd, cmd, ptr, ilen, olen)
    int fd, cmd, ilen, olen;
    char *ptr;
{
    struct strioctl str;

    str.ic_cmd = cmd;
    str.ic_timout = 0;
    str.ic_len = ilen;
    str.ic_dp = ptr;
    if (ioctl(fd, I_STR, &str) == -1)
	return -1;
    if (str.ic_len != olen)
	fprintf(stderr, "strioctl: expected %d bytes, got %d for cmd %x\n",
	       olen, str.ic_len, cmd);
    return 0;
}

static void
get_ppp_stats(curp)
    struct ppp_stats *curp;
{
    if (strioctl(fd_info, PPPIO_GETSTAT, curp, 0, sizeof(*curp)) < 0) {
	fprintf(stderr, "%s: ", ProgramName);
	if (errno == EINVAL)
	    fprintf(stderr, "kernel support missing\n");
	else
	    perror("couldn't get PPP statistics");
	exit(1);
    }
}

#endif /* STREAMS */

#define MAX0(a)		((int)(a) > 0? (a): 0)
#define V(offset)	MAX0(cur.offset - old.offset)

void Init()
{
  Now=time(0L);
  get_ppp_stats(&cur);
}


void GetPPPActivityCB( w, closure, call_data )
Widget	w;
XtPointer closure;
XtPointer call_data;	/* pointer to (double) return value */
{
  int			cps=0;
  int			LastState=State;
  struct ppp_stats	old=cur;
  time_t		t=time(NULL);
  struct tm		*tm=localtime(&t);
  Arg			args[1];
  char			s[120];
  time_t		Last=Now;
  int			DeltaT;

  Now=time(0L);
  DeltaT=Now-Last;

  State = get_active(); 

  if (State!=LastState)			/* Zustandswechsel ...		    */
    {
      switch(LastState)			/* den aktuellen Zustand verlassen  */
	{
	case STATE_OFFLINE:
	  break;
	case STATE_ONLINE:
	  OnlineTill=Now;
	  if (resources.show_label)
	    {
	      struct tm	Since;
	      struct tm	Till;
	      int SecsRunning=OnlineTill - OnlineSince;

	      memcpy(&Since, localtime(&OnlineSince), sizeof(Since));
	      memcpy(&Till, localtime(&OnlineTill), sizeof(Till));

	      snprintf(s, sizeof(s)-1, "offline %d:%02d (%d:%02d-%d:%02d) %dcps %dKB",
		       SecsRunning/60, SecsRunning%60,
		       Since.tm_hour, Since.tm_min, Till.tm_hour, Till.tm_min,
		       SecsRunning ? ByteCount/SecsRunning : 0,
		       ByteCount/1024);

	      XtSetArg (args[0], XtNlabel, s);
	      XtSetValues (label_wid, args, ONE);
	    }
	  break;
	case STATE_ACTIVE:
	  break;
	case STATE_TRYING:
	  break;
	}

      switch(State)			/* in den neuen Zustand wechseln    */
	{
	case STATE_OFFLINE:
	  XtSetArg(args[0], XtNbackground, bgcolor);
	  XtSetValues(w, args, 1);
	  break;
	case STATE_ONLINE:
	  ByteCount=0;
	  OnlineSince=Now;
	  XtSetArg(args[0], XtNbackground, onlinecolor);
	  XtSetValues(w, args, 1);
	  break;
	case STATE_ACTIVE:
	  XtSetArg(args[0], XtNbackground, activecolor );
	  XtSetValues(w, args, 1);
	  break;
	case STATE_TRYING:
	  XtSetArg(args[0], XtNbackground, tryingcolor );
	  XtSetValues(w, args, 1);
	  break;
	}
    }

  switch(State)				/* Actions performed while in a	    */
    {					/* specific State		    */
    case STATE_OFFLINE:
      break;
    case STATE_ONLINE:
      {
	int		SecsRunning=Now - OnlineSince;
	struct tm	*Since=localtime(&OnlineSince);
	int		AdditionalBytes;

	get_ppp_stats(&cur);
	AdditionalBytes  = V(p.ppp_ibytes)+V(p.ppp_obytes);
	ByteCount+=AdditionalBytes;

	if (DeltaT>0)
	  cps = AdditionalBytes / DeltaT;
	if (cps < 0) 
	  cps = 0;

	snprintf(s, sizeof(s)-1, "online %d:%02d (%d:%02d-...) %5dcps %dKB",
		 SecsRunning/60, SecsRunning%60,
		 Since->tm_hour, Since->tm_min,
		 cps, ByteCount/1024);
	
	XtSetArg (args[0], XtNlabel, s);
	XtSetValues (label_wid, args, ONE);
      }
      break;
    case STATE_ACTIVE:
      break;
    case STATE_TRYING:
      break;
    }

  *((double *)call_data) = (double)cps / 8000.0;	/* unit: 8000Bytes/sec */
}

void ToggleActive(Widget w, XtPointer p, XEvent *e, Boolean *c)
{
  if (e->type == ButtonPress)
    {
      if (State==STATE_OFFLINE)
	system(resources.activate);
      else
	system(resources.deactivate);
    }
} /* ToggleActive() */

int main(argc, argv)
     int argc;
     char **argv;
{
  XtAppContext app_con;
  Widget toplevel, load, pane, load_parent;
  Pixmap icon_pixmap = None;
  XrmValue namein, pixelout;
  Arg args[1];

  ProgramName = argv[0];
    
  toplevel = XtAppInitialize(&app_con, "XPPPLoad", options,
			     XtNumber(options),
			     &argc, argv, NULL, NULL, (Cardinal) 0);

  XtGetApplicationResources( toplevel, (XtPointer) &resources,
			     my_resources, XtNumber(my_resources),
			     NULL, (Cardinal) 0);
  
  if (argc != 2) usage();
  IfName=XtNewString(argv[1]);

#ifndef STREAMS
  {
    struct ifreq ifr;
    
    fd_info = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd_info < 0) {
      fprintf(stderr, "%s: ", ProgramName);
      perror("couldn't create IP socket");
      exit(1);
    }

#ifdef linux
#undef  ifr_name
#define ifr_name ifr_ifrn.ifrn_name
#endif
    strncpy(ifr.ifr_name, IfName, sizeof(ifr.ifr_name));
    if (ioctl(fd_info, SIOCGIFFLAGS, (caddr_t)&ifr) < 0)
      if ( errno != ENODEV )
	{
	  perror(IfName);
	  exit(1);
	}
  }
#else	/* STREAMS */
  {
#ifdef __osf__
    const char *dev = "/dev/streams/ppp";
#else
    const char *dev = "/dev/ppp";
#endif
    int unit=0;

    if (sscanf(IfName, "ppp%d", &unit) != 1)
      {
	fprintf(stderr, "%s: invalid interface '%s' specified\n",
		ProgramName, IfName);
      }

    if ((fd_info = open(dev, O_RDONLY)) < 0)
      {
	fprintf(stderr, "%s: couldn't open ", ProgramName);
	perror(dev);
	exit(1);
      }
    if (strioctl(fd_info, PPPIO_ATTACH, &unit, sizeof(int), 0) < 0)
      {
	fprintf(stderr, "%s: ppp%d is not available\n", ProgramName, unit);
	exit(1);
      }
  }
#endif	/* STREAMS */
  Init();

    
  /*
   * This is a hack so that f.delete will do something useful in this
   * single-window application.
   */
  XtAppAddActions (app_con, xpppload_actions, XtNumber(xpppload_actions));
  XtOverrideTranslations(toplevel,
			 XtParseTranslationTable ("<Message>WM_PROTOCOLS: quit()"));

  XtSetArg (args[0], XtNiconPixmap, &icon_pixmap);
  XtGetValues(toplevel, args, ONE);
  if (icon_pixmap == None)
    {
      XtSetArg(args[0], XtNiconPixmap,
	       XCreateBitmapFromData(XtDisplay(toplevel),
				     XtScreen(toplevel)->root,
				     (char *)xpppload_bits,
				     xpppload_width, xpppload_height));
      XtSetValues (toplevel, args, ONE);
    }
    
  if (resources.show_label)
    {
      pane = XtCreateManagedWidget ("paned", panedWidgetClass,
				    toplevel, NULL, ZERO);
      
      XtSetArg (args[0], XtNlabel, "uninitialized");
      label_wid = XtCreateManagedWidget ("label", labelWidgetClass,
					 pane, args, 1);
      load_parent = pane;
    }
  else
    load_parent = toplevel;

  load = XtCreateManagedWidget ("load", stripChartWidgetClass,
				load_parent, NULL, ZERO);

  XtSetArg (args[0], XtNbackground, &bgcolor);
  XtGetValues(toplevel, args, ONE);
    
  if (resources.online_color)
    {
      namein.addr = resources.online_color;
      namein.size = strlen(resources.online_color) + 1;
      XtConvert(load, XtRString, &namein, XtRPixel, &pixelout);
      if (!pixelout.addr) {
	fprintf(stderr, "could not convert online color `%s'.\n",
		resources.online_color);
        exit(1);
      }
      onlinecolor = *(Pixel*)(pixelout.addr);
    }
  else 
    onlinecolor = bgcolor;

  if (resources.trying_color)
    {
      namein.addr = resources.trying_color;
      namein.size = strlen(resources.trying_color) + 1;
      XtConvert(load, XtRString, &namein, XtRPixel, &pixelout);
      if (!pixelout.addr)
	{
	  fprintf(stderr, "could not convert trying color `%s'.\n",
		  resources.trying_color);
	  exit(1);
	}
      tryingcolor = *(Pixel*)(pixelout.addr);
    }
  else
    tryingcolor = bgcolor;

  if (resources.active_color)
    {
      namein.addr = resources.active_color;
      namein.size = strlen(resources.active_color) + 1;
      XtConvert(load, XtRString, &namein, XtRPixel, &pixelout);
      if (!pixelout.addr)
	{
	  fprintf(stderr, "could not convert active color `%s'.\n",
		  resources.online_color);
	  exit(1);
	}
      activecolor = *(Pixel*)(pixelout.addr);
    } 
  else 
    activecolor = bgcolor;

  XtAddCallback(load, XtNgetValue, GetPPPActivityCB, NULL);

  XtAddEventHandler(toplevel, ButtonPressMask, ButtonPressMask,
		    ToggleActive, NULL);
    
  XtRealizeWidget (toplevel);

  wm_delete_window = XInternAtom (XtDisplay(toplevel), "WM_DELETE_WINDOW",
				  False);
  (void) XSetWMProtocols (XtDisplay(toplevel), XtWindow(toplevel),
			  &wm_delete_window, 1);

  XtAppMainLoop(app_con);

  return 0;
}

static void quit (w, event, params, num_params)
    Widget w;
    XEvent *event;
    String *params;
    Cardinal *num_params;
{
    if (event->type == ClientMessage &&
        event->xclient.data.l[0] != wm_delete_window) {
        XBell (XtDisplay(w), 0);
        return;
    }
    XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
    exit (0);
} /* main() */
